package com.jclark.xml.parse;

import java.io.IOException;
import java.io.InputStream;
import java.io.Writer;
import java.net.URL;
import java.util.Enumeration;
import java.util.ResourceBundle;
import java.util.MissingResourceException;
import java.util.Locale;
import java.text.MessageFormat;

import com.jclark.util.Hashtable;

import com.jclark.xml.tok.*;

class EntityParser extends ContentToken implements StartElementEvent, EndElementEvent, CharacterDataEvent, ProcessingInstructionEvent, EndPrologEvent {
  private static class Entity {
    byte[] text;
    String systemId;
    String publicId;
    URL baseURL;
    String notationName;
    boolean open;
    boolean referenced;
  }

  private static class DTDImpl implements DTD {
    Hashtable elementTypeTable = new Hashtable();
    Hashtable generalEntityTable = new Hashtable();
    Hashtable paramEntityTable = new Hashtable();
    Hashtable notationTable = new Hashtable();
    boolean complete = true;
    boolean standalone = false;

    public boolean isComplete() {
      return complete;
    }

    public boolean isStandalone() {
      return standalone;
    }

    public Enumeration entityNames() {
      return generalEntityTable.keys();
    }

    public Enumeration notationNames() {
      return notationTable.keys();
    }

    public String getNotationSystemId(String notationName) {
      Entity entity = (Entity)notationTable.get(notationName);
      if (entity != null)
	return entity.systemId;
      return null;
    }

    public String getNotationPublicId(String notationName) {
      Entity entity = (Entity)notationTable.get(notationName);
      if (entity != null)
	return entity.publicId;
      return null;
    }

    public URL getNotationBase(String notationName) {
      Entity entity = (Entity)notationTable.get(notationName);
      if (entity != null)
	return entity.baseURL;
      return null;
    }

    public String getEntitySystemId(String entityName) {
      Entity entity = (Entity)generalEntityTable.get(entityName);
      if (entity != null)
	return entity.systemId;
      return null;
    }

    public String getEntityPublicId(String entityName) {
      Entity entity = (Entity)generalEntityTable.get(entityName);
      if (entity != null)
	return entity.publicId;
      return null;
    }

    public String getEntityNotationName(String entityName) {
      Entity entity = (Entity)generalEntityTable.get(entityName);
      if (entity != null)
	return entity.notationName;
      return null;
    }

    public String getEntityReplacementText(String entityName) {
      Entity entity = (Entity)generalEntityTable.get(entityName);
      if (entity == null || entity.text == null)
	return null;
      StringBuffer buf = new StringBuffer(entity.text.length >> 1);
      for (int i = 0; i < entity.text.length; i += 2)
	buf.append((char)(((entity.text[i] & 0xFF) << 8)
			  | (entity.text[i + 1] & 0xFF)));
      return buf.toString();
    }

    public boolean getElementTypeElementContent(String name) {
      ElementType elementType = (ElementType)elementTypeTable.get(name);
      if (elementType != null)
	return elementType.isElementContent();
      return false;
    }

    public URL getEntityBase(String entityName) {
      Entity entity = (Entity)generalEntityTable.get(entityName);
      if (entity != null)
	return entity.baseURL;
      return null;
    }
  }

  private static final boolean forceStandalone = false;

  private static final int READSIZE = 1024*8;

  private EntityParser parent;
  private String internalEntityName;
  private boolean isParameterEntity;
  private byte[] buf;
  private int bufStart;
  private int bufEnd;
  private int currentTokenStart;
  private InputStream in;
  private URL baseURL;
  private String location;
  private Position pos = new Position();
  // The offset in buffer corresponding to pos.
  private int posOff = 0;
  private long bufEndStreamOffset = 0;
  private Encoding enc;
  // True if the encoding in the XML declaration should be ignored.
  private boolean ignoreDeclEnc;
  private /* final */ int minBPC;
  private int fixBPC;
  private StringConversionCache stringCache;
  private Encoding internalEnc;
  private StringConversionCache internalStringCache;
  private Application app;
  private DTDImpl dtd;
  private EntityManager entityManager;
  private Locale locale;

  private int nameStart;
  // Some temporary buffers
  private Buffer valueBuf;
  private char[] data;
  private static final int INIT_DATA_BUF_SIZE = 65;
  private int dataLength;
  private char[] dataBuf;
  private String[] attValues;
  private String[] attNames;
  private int nAttributes;
  private boolean[] defaultSpecified;

  EntityParser(OpenEntity entity, EntityManager entityManager, Application app,
	       Locale locale, EntityParser parent) throws IOException {
    this.in = entity.getInputStream();
    this.app = app;
    this.locale = locale;
    this.baseURL = entity.getBase();
    this.location = entity.getLocation();
    this.entityManager = entityManager;
    buf = new byte[READSIZE * 2];
    currentTokenStart = bufStart = bufEnd = 0;
    while (bufEnd - bufStart < 4 && fill())
      ;
    enc = Encoding.getInitialEncoding(buf, bufStart, bufEnd, this);
    bufStart = getTokenEnd();
    posOff = bufStart; // ignore the byte order mark in computing columns
    if (enc == null)
      fatal(MessageId.BAD_INITIAL_BYTES);
    minBPC = enc.getMinBytesPerChar();
    fixBPC = enc.getFixedBytesPerChar();
    String encName = entity.getEncoding();
    if (encName != null) {
      ignoreDeclEnc = true;
      enc = enc.getEncoding(encName);
      if (enc == null)
	fatal(MessageId.UNSUPPORTED_ENCODING);
      if (enc.getMinBytesPerChar() != minBPC)
	fatal(MessageId.BAD_EXTERN_ENCODING);
      fixBPC = enc.getFixedBytesPerChar();
    }
    stringCache = new StringConversionCache(enc);
    valueBuf = new Buffer();
    dataBuf = new char[INIT_DATA_BUF_SIZE];
    internalEnc = Encoding.getInternalEncoding();
    internalStringCache = new StringConversionCache(internalEnc);
    if (parent != null)
      dtd = parent.dtd;
    else
      dtd = new DTDImpl();
  }

  private EntityParser(byte[] buf, String entityName, boolean isParameterEntity, EntityParser parent) {
    this.internalEntityName = entityName;
    this.isParameterEntity = isParameterEntity;
    this.buf = buf;
    this.parent = parent;
    baseURL = parent.baseURL;
    entityManager = parent.entityManager;
    currentTokenStart = bufStart = 0;
    bufEnd = buf.length;
    app = parent.app;
    locale = parent.locale;
    enc = internalEnc = parent.internalEnc;
    stringCache = internalStringCache = parent.internalStringCache;
    minBPC = enc.getMinBytesPerChar();
    fixBPC = enc.getFixedBytesPerChar();
    dtd = parent.dtd;
    valueBuf = parent.valueBuf;
    dataBuf = parent.dataBuf;
  }

  void parseDocumentEntity() throws IOException {
    app.startDocument();
    parseDecls(PrologParser.PROLOG);
    parseContent(true);
    parseMisc();
    app.endDocument();
  }

  private void parseExternalTextEntity() throws IOException {
    for (;;) {
      try {
	if (enc.tokenizeContent(buf, bufStart, bufEnd, this)
	    == Encoding.TOK_XML_DECL) {
	  currentTokenStart = bufStart;
	  bufStart = getTokenEnd();
	  handleXmlDecl(true);
	}
	break;
      }
      catch (InvalidTokenException e) {
	break;
      }
      catch (TokenException e) {
	if (!fill())
	  break;
      }
    }
    parseContent(false);
  }

  private void handleXmlDecl(boolean notDocumentEntity) throws NotWellFormedException {
    try {
      TextDecl textDecl;
      if (notDocumentEntity)
	textDecl = new TextDecl(enc, buf, currentTokenStart, bufStart);
      else {
	XmlDecl xmlDecl = new XmlDecl(enc, buf, currentTokenStart, bufStart);
	dtd.standalone = xmlDecl.isStandalone();
	textDecl = xmlDecl;
      }
      if (!ignoreDeclEnc) {
	enc = enc.getEncoding(textDecl.getEncoding());
	if (enc == null)
	  fatal(MessageId.UNSUPPORTED_ENCODING);
	if (enc.getMinBytesPerChar() != minBPC)
	  fatal(MessageId.BAD_DECL_ENCODING);
	stringCache.setEncoding(enc);
	fixBPC = enc.getFixedBytesPerChar();
      }
    }
    catch (InvalidTokenException e) {
      currentTokenStart = e.getOffset();
      fatal(MessageId.INVALID_XML_DECLARATION);
    }
  }

  static class DeclState {
    Entity entity;
    ElementType elementType;
    String attributeName;
    boolean attributeIsCdata;
    String doctypeSystemId;
    String doctypePublicId;
  }

  private void parseDecls(byte type) throws IOException {
    PrologParser pp = new PrologParser(type);
    DeclState declState = new DeclState();
    try {
      for (;;) {
	int tok;
	try {
	  tok = tokenizeProlog();
	}
	catch (EndOfPrologException e) {
	  if (type != PrologParser.PROLOG)
	    fatal(MessageId.SYNTAX_ERROR);
	  pp.end();
	  break;
	}
	catch (EmptyTokenException e) {
	  if (type == PrologParser.PROLOG) {
	    currentTokenStart = bufStart;
	    fatal(MessageId.NO_DOCUMENT_ELEMENT);
	  }
	  pp.end();
	  break;
	}
	prologAction(pp.action(tok, buf, currentTokenStart, bufStart, enc),
		     pp,
		     declState);
      }
    }
    catch (PrologSyntaxException e) {
      fatal(MessageId.SYNTAX_ERROR);
    }
    if (declState.doctypeSystemId != null) {
      if (!dtd.standalone && !forceStandalone) {
	OpenEntity openEntity
	  = entityManager.open(declState.doctypeSystemId,
			       baseURL,
			       declState.doctypePublicId);
	if (openEntity != null) {
	  new EntityParser(openEntity, entityManager, app, locale, this)
	    .parseDecls(PrologParser.EXTERNAL_ENTITY);
	  return;
	}
      }
      dtd.complete = false;
    }
    if (type == PrologParser.PROLOG)
      app.endProlog(this);
  }

  void parseInnerParamEntity(PrologParser pp, DeclState declState) throws IOException {
    int groupLevel = pp.getGroupLevel();
    try {
      for (;;) {
	int tok = tokenizeProlog();
	prologAction(pp.action(tok, buf, currentTokenStart, bufStart, enc),
		     pp,
		     declState);
	if (tok == Encoding.TOK_DECL_CLOSE)
	  fatal(MessageId.PE_DECL_NESTING);
      }
    }
    catch (EndOfPrologException e) {
      fatal(MessageId.SYNTAX_ERROR);
    }
    catch (PrologSyntaxException e) {
      fatal(MessageId.SYNTAX_ERROR);
    }
    catch (EmptyTokenException e) { }
    if (pp.getGroupLevel() != groupLevel)
      fatal(MessageId.PE_GROUP_NESTING);
  }

  void prologAction(int action, PrologParser pp, DeclState declState) throws IOException {
    String name;
    switch (action) {
    case PrologParser.ACTION_XML_DECL:
      handleXmlDecl(false);
      break;
    case PrologParser.ACTION_TEXT_DECL:
      handleXmlDecl(true);
      break;
    case PrologParser.ACTION_ENTITY_PUBLIC_ID:
    case PrologParser.ACTION_DOCTYPE_PUBLIC_ID:
    case PrologParser.ACTION_NOTATION_PUBLIC_ID:
      try {
	String id = enc.getPublicId(buf, currentTokenStart, bufStart);
	if (action == PrologParser.ACTION_DOCTYPE_PUBLIC_ID)
	  declState.doctypePublicId = id;
	else if (declState.entity != null)
	  declState.entity.publicId = id;
      }
      catch (InvalidTokenException e) {
	currentTokenStart = e.getOffset();
	fatal(MessageId.PUBID_CHAR);
      }
      break;
    case PrologParser.ACTION_NOTATION_NAME:
      startEntityDecl(dtd.notationTable, declState);
      break;
    case PrologParser.ACTION_GENERAL_ENTITY_NAME:
      startEntityDecl(dtd.generalEntityTable, declState);
      break;
    case PrologParser.ACTION_PARAM_ENTITY_NAME:
      startEntityDecl(dtd.paramEntityTable, declState);
      break;
    case PrologParser.ACTION_ENTITY_VALUE_NO_PEREFS:
    case PrologParser.ACTION_ENTITY_VALUE_WITH_PEREFS:
      byte[] text = makeReplacementText(action == PrologParser.ACTION_ENTITY_VALUE_WITH_PEREFS);
      if (declState.entity != null)
	declState.entity.text = text;
      break;
    case PrologParser.ACTION_NOTATION_SYSTEM_ID:
    case PrologParser.ACTION_ENTITY_SYSTEM_ID:
      if (declState.entity != null) {
	declState.entity.systemId
	  = stringCache.convert(buf,
				currentTokenStart + minBPC,
				bufStart - minBPC,
				false);
	declState.entity.baseURL = baseURL;
      }
      break;
    case PrologParser.ACTION_ENTITY_NOTATION_NAME:
      if (declState.entity != null)
	declState.entity.notationName
	  = stringCache.convert(buf, currentTokenStart, bufStart, true);
      break;
    case PrologParser.ACTION_DOCTYPE_SYSTEM_ID:
      // Don't set dtd.complete to false here, because it isn't incomplete
      // as far as processing of the internal subset is concerned.
      declState.doctypeSystemId
	  = stringCache.convert(buf,
				currentTokenStart + minBPC,
				bufStart - minBPC,
				true);
      break;
    case PrologParser.ACTION_INNER_PARAM_ENTITY_REF:
    case PrologParser.ACTION_OUTER_PARAM_ENTITY_REF:
      name = stringCache.convert(buf,
				 currentTokenStart + minBPC,
				 bufStart - minBPC,
				 true);
      Entity entity = (Entity)dtd.paramEntityTable.get(name);
      if (entity == null) {
	if (dtd.complete)
	  fatal(MessageId.UNDEF_PEREF, name);
	break;
      }
      EntityParser parser = makeParserForEntity(entity, name, true);
      if (parser == null || dtd.standalone || forceStandalone) {
	dtd.complete = false;
	break;
      }
      entity.open = true;
      if (action == PrologParser.ACTION_OUTER_PARAM_ENTITY_REF)
	parser.parseDecls(entity.text != null
			  ? PrologParser.INTERNAL_ENTITY
			  : PrologParser.EXTERNAL_ENTITY);
      else
	parser.parseInnerParamEntity(pp, declState);
      entity.open = false;
      break;
      /* Default attribute processing. */
    case PrologParser.ACTION_ATTLIST_ELEMENT_NAME:
      String gi = stringCache.convert(buf,
				      currentTokenStart,
				      bufStart,
				      true);
      declState.elementType = (ElementType)dtd.elementTypeTable.get(gi);
      if (declState.elementType == null) {
	declState.elementType = new ElementType();
	dtd.elementTypeTable.put(gi, declState.elementType);
      }
      break;
    case PrologParser.ACTION_ATTRIBUTE_NAME:
      declState.attributeName = stringCache.convert(buf,
						    currentTokenStart,
						    bufStart,
						    true);
      declState.attributeIsCdata = false;
      break;
    case PrologParser.ACTION_ATTRIBUTE_TYPE_CDATA:
      declState.attributeIsCdata = true;
      break;
    case PrologParser.ACTION_IMPLIED_ATTRIBUTE_VALUE:
    case PrologParser.ACTION_REQUIRED_ATTRIBUTE_VALUE:
      declState.elementType
	       .appendAttribute(declState.attributeName,
				null,
				declState.attributeIsCdata);
      break;
    case PrologParser.ACTION_DEFAULT_ATTRIBUTE_VALUE:
    case PrologParser.ACTION_FIXED_ATTRIBUTE_VALUE:
      declState.elementType
	       .appendAttribute(declState.attributeName,
				makeAttributeValue(declState.attributeIsCdata,
						   buf,
						   currentTokenStart + minBPC,
						   bufStart - minBPC),
				declState.attributeIsCdata);
      break;
    case PrologParser.ACTION_PI:
      nameStart = currentTokenStart + minBPC + minBPC;
      app.processingInstruction(this);
      break;
    case PrologParser.ACTION_IGNORE_SECT:
      skipIgnoreSect();
      break;
    case PrologParser.ACTION_ELEMENT_NAME:
      name = stringCache.convert(buf, currentTokenStart, bufStart, true);
      declState.elementType = (ElementType)dtd.elementTypeTable.get(name);
      if (declState.elementType == null) {
	declState.elementType = new ElementType();
	dtd.elementTypeTable.put(name, declState.elementType);
      }
      declState.elementType.setElementContent(true);
      break;
    case PrologParser.ACTION_CONTENT_ANY:
    case PrologParser.ACTION_CONTENT_PCDATA:
      declState.elementType.setElementContent(false);
      break;
    }
  }

  private final void startEntityDecl(Hashtable table, DeclState declState) {
    String name = stringCache.convert(buf, currentTokenStart, bufStart, true);
    declState.entity = (Entity)table.get(name);
    if (declState.entity == null) {
      declState.entity = new Entity();
      table.put(name, declState.entity);
    }
    else
      declState.entity = null;
  }

  private final void skipIgnoreSect() throws IOException {
    for (;;) {
      try {
	bufStart = enc.skipIgnoreSect(buf, bufStart, bufEnd);
	return;
      }
      catch (PartialTokenException e) {
	if (!fill()) {
	  currentTokenStart = bufStart;
	  fatal(MessageId.UNCLOSED_CONDITIONAL_SECTION);
	}
      }
      catch (InvalidTokenException e) {
	currentTokenStart = e.getOffset();
	fatal(MessageId.IGNORE_SECT_CHAR);
      }
    }
  }

  private final void parseContent(boolean oneElement) throws IOException {
    byte[] buf = this.buf;
    int bufEnd = this.bufEnd;
    int bufStart = this.bufStart;
    Encoding enc = this.enc;
    int nOpenElements = 0;
    byte[] openElementNameBuf = new byte[64];
    // Indexed by nOpenElements
    int[] openElementNameStart = new int[8];
    openElementNameStart[0] = 0;
    for (;;) {
      try {
	switch (enc.tokenizeContent(buf, bufStart, bufEnd, this)) {
	case Encoding.TOK_START_TAG_WITH_ATTS:
	  storeAtts();
	  /* fall through */
	case Encoding.TOK_START_TAG_NO_ATTS:
	  if (nOpenElements + 1 >= openElementNameStart.length)
	    openElementNameStart = grow(openElementNameStart);
	  nameStart = bufStart + minBPC;
	  nAttributes = -1;
	  app.startElement(this);
	  int nameLength = getNameEnd() - nameStart;
	  int nameBufEnd = openElementNameStart[nOpenElements];
	  if ((openElementNameStart[nOpenElements + 1]
	       = nameBufEnd + nameLength)
	      > openElementNameBuf.length) {
	    byte[] tem
	      = new byte[(openElementNameBuf.length << 1) + nameLength];
	    System.arraycopy(openElementNameBuf,
			     0,
			     tem,
			     0,
			     openElementNameStart[nOpenElements]);
	    openElementNameBuf = tem;
	  }
	  copyBytes(buf, nameStart, openElementNameBuf, nameBufEnd, nameLength);
	  nOpenElements++;
	  break;
	case Encoding.TOK_EMPTY_ELEMENT_WITH_ATTS:
	  storeAtts();
	  /* fall through */
	case Encoding.TOK_EMPTY_ELEMENT_NO_ATTS:
	  nameStart = bufStart + minBPC;
	  nAttributes = -1;
	  app.startElement(this);
	  app.endElement(this);
	  if (oneElement && nOpenElements == 0) {
	    this.bufStart = getTokenEnd();
	    return;
	  }
	  break;
	case Encoding.TOK_END_TAG:
	  if (nOpenElements == 0) {
	    currentTokenStart = bufStart;
	    fatal(MessageId.INVALID_END_TAG);
	  }
	  --nOpenElements;
	  nameStart = bufStart + 2*minBPC;
	  if (!bytesEqual(openElementNameBuf,
			  openElementNameStart[nOpenElements],
			  openElementNameStart[nOpenElements + 1],
			  buf,
			  nameStart,
			  getNameEnd())) {
	    String expected
	      = stringCache.convert(openElementNameBuf,
				    openElementNameStart[nOpenElements],
				    openElementNameStart[nOpenElements + 1],
				    false);
	    String got = stringCache.convert(buf,
					     nameStart,
					     getNameEnd(),
					     false);
	    currentTokenStart = bufStart;
	    fatal(MessageId.MISMATCHED_END_TAG, got, expected);
	  }
	  app.endElement(this);
	  if (oneElement && nOpenElements == 0) {
	    this.bufStart = getTokenEnd();
	    return;
	  }
	  break;
	case Encoding.TOK_DATA_CHARS:
	  data = null;
	  this.bufStart = bufStart;
	  app.characterData(this);
	  break;
	case Encoding.TOK_DATA_NEWLINE:
	  dataBuf[0] = '\n';
	  dataLength = 1;
	  data = dataBuf;
	  app.characterData(this);
	  break;
	case Encoding.TOK_MAGIC_ENTITY_REF:
	case Encoding.TOK_CHAR_REF:
	  dataBuf[0] = getRefChar();
	  dataLength = 1;
	  data = dataBuf;
	  app.characterData(this);
	  break;
	case Encoding.TOK_CHAR_PAIR_REF:
	  getRefCharPair(dataBuf, 0);
	  data = dataBuf;
	  dataLength = 2;
	  app.characterData(this);
	  break;
	case Encoding.TOK_CDATA_SECTION:
	  convertData(bufStart + 9*minBPC, getTokenEnd() - 3*minBPC);
	  app.characterData(this);
	  break;
	case Encoding.TOK_ENTITY_REF:
	  {
	    String name = stringCache.convert(buf,
					      bufStart + minBPC,
					      getTokenEnd() - minBPC,
					      true);
	    Entity entity = (Entity)dtd.generalEntityTable.get(name);
	    if (entity == null) {
	      if (dtd.complete || dtd.standalone) {
		currentTokenStart = bufStart;
		fatal(MessageId.UNDEF_REF, name);
	      }
	      break;
	    }
	    EntityParser parser = makeParserForEntity(entity, name, false);
	    if (parser == null)
	      break;
	    entity.open = true;
	    if (entity.text != null) {
	      currentTokenStart = this.bufStart = bufStart;
	      parser.parseContent(false);
	    }
	    else
	      parser.parseExternalTextEntity();
	    entity.open = false;
	    break;
	  }
	case Encoding.TOK_PI:
	  nameStart = bufStart + minBPC*2;
	  app.processingInstruction(this);
	  break;
	case Encoding.TOK_XML_DECL:
	  currentTokenStart = bufStart;
	  fatal(MessageId.MISPLACED_XML_DECL);
	}
	bufStart = getTokenEnd();
      }
      catch (EmptyTokenException e) {
	this.bufStart = bufStart;
	if (!fill()) {
	  if (oneElement || nOpenElements > 0) {
	    currentTokenStart = this.bufStart;
	    fatal(MessageId.MISSING_END_TAG);
	  }
	  return;
	}
	buf = this.buf;
	bufStart = this.bufStart;
	bufEnd = this.bufEnd;
      }
      catch (PartialTokenException e) {
	this.bufStart = bufStart;
	if (!fill()) {
	  currentTokenStart = this.bufStart;
	  fatal(MessageId.UNCLOSED_TOKEN);
	}
	buf = this.buf;
	bufStart = this.bufStart;
	bufEnd = this.bufEnd;
      }
      catch (ExtensibleTokenException e) {
	this.bufStart = bufStart;
	if (!fill()) {
	  if (oneElement || nOpenElements > 0) {
	    currentTokenStart = this.bufStart;
	    fatal(MessageId.MISSING_END_TAG);
	  }
	  dataBuf[0] = '\n';
	  dataLength = 1;
	  data = dataBuf;
	  app.characterData(this);
	  return;
	}
	buf = this.buf;
	bufStart = this.bufStart;
	bufEnd = this.bufEnd;
      }
      catch (InvalidTokenException e) {
	currentTokenStart = e.getOffset();
	reportInvalidToken(e);
      }
    }
  }

  private EntityParser makeParserForEntity(Entity entity, String name, boolean isParameter) throws IOException {
    entity.referenced = true;
    if (entity.open)
      fatal(MessageId.RECURSION);
    if (entity.notationName != null)
      fatal(MessageId.UNPARSED_REF);
    if (entity.text != null)
      return new EntityParser(entity.text, name, isParameter, this);
    OpenEntity openEntity
      = entityManager.open(entity.systemId, entity.baseURL, entity.publicId);
    if (openEntity == null)
      return null;
    return new EntityParser(openEntity, entityManager, app, locale, this);
  }

  private final void storeAtts() throws NotWellFormedException {
    int i = getAttributeSpecifiedCount();
    ElementType elementType = null;
    boolean gotElementType = false;
    while (i != 0) {
      --i;
      if (!isAttributeNormalized(i)) {
	valueBuf.clear();
	if (!gotElementType)
	  elementType
	    = (ElementType)dtd.elementTypeTable
	                      .get(stringCache.convert(buf,
						       bufStart + minBPC,
						       getNameEnd(),
						       true));
	boolean isCdata;
	if (elementType != null) {
	  String attName
	    = stringCache.convert(buf,
				  getAttributeNameStart(i),
				  getAttributeNameEnd(i),
				  true);
	  isCdata = elementType.isAttributeCdata(attName);
	}
	else
	  isCdata = true;
	String val = makeAttributeValue(isCdata,
					buf,
					getAttributeValueStart(i),
					getAttributeValueEnd(i));
	setAttributeValue(i, val);
      }
    }
  }

  private String makeAttributeValue(boolean isCdata,
				    byte[] buf,
				    int start,
				    int end) throws NotWellFormedException {
 				      
    /* appendAttributeValue will trash currentTokenStart. */
    int saveCurrentTokenStart = currentTokenStart;
    valueBuf.clear();
    appendAttributeValue(isCdata, start, end, valueBuf);
    if (!isCdata
	&& valueBuf.length() > 0
	&& valueBuf.charAt(valueBuf.length() - 1) == ' ')
      valueBuf.chop();
    currentTokenStart = saveCurrentTokenStart;
    return valueBuf.toString();
  }

  private void appendAttributeValue(boolean isCdata,
				    int start,
				    int end,
				    Buffer valueBuf) throws NotWellFormedException {
    Token t = new Token();
    try {
      for (;;) {
	int tok;
	int nextStart;
	try {
	  tok = enc.tokenizeAttributeValue(buf, start, end, t);
	  nextStart = t.getTokenEnd();
	}
	catch (ExtensibleTokenException e) {
	  tok = e.getTokenType();
	  nextStart = end;
	}
	currentTokenStart = start;
	switch (tok) {
	case Encoding.TOK_ATTRIBUTE_VALUE_S:
	  if (!isCdata) {
	    if (valueBuf.length() > 0
		&& valueBuf.charAt(valueBuf.length() - 1) != ' ')
	      valueBuf.append(' ');
	    break;
	  }
	  /* fall through */
	case Encoding.TOK_DATA_CHARS:
	  valueBuf.append(enc, buf, start, t.getTokenEnd());
	  break;
	case Encoding.TOK_MAGIC_ENTITY_REF:
	case Encoding.TOK_CHAR_REF:
	  valueBuf.append(t.getRefChar());
	  break;
	case Encoding.TOK_CHAR_PAIR_REF:
	  valueBuf.appendRefCharPair(t);
	  break;
	case Encoding.TOK_DATA_NEWLINE:
	  if (isCdata
	      || (valueBuf.length() > 0
		  && valueBuf.charAt(valueBuf.length() - 1) != ' '))
	    valueBuf.append(' ');
	  break;
	case Encoding.TOK_ENTITY_REF:
	  String name = stringCache.convert(buf,
					    start + minBPC,
					    t.getTokenEnd() - minBPC,
					    true);
	  Entity entity = (Entity)dtd.generalEntityTable.get(name);
	  if (entity == null) {
	    if (dtd.complete || dtd.standalone)
	      fatal(MessageId.UNDEF_REF, name);
	    break;
	  }
	  if (entity.systemId != null)
	    fatal(MessageId.EXTERN_REF_ATTVAL);
	  try {
	    EntityParser parser = makeParserForEntity(entity, name, false);
	    entity.open = true;
	    parser.appendAttributeValue(isCdata, 0, parser.bufEnd, valueBuf);
	    entity.open = false;
	  }
	  catch (NotWellFormedException e) {
	    throw e;
	  }
	  catch (IOException e) {
	    // Shouldn't happen since the entity is internal.
	    throw new Error("unexpected IOException");
	  }
	  break;
	default:
	  throw new Error("attribute value botch");
	}
	start = nextStart;
      }
    }
    catch (PartialTokenException e) {
      currentTokenStart = end;
      fatal(MessageId.NOT_WELL_FORMED);
    }
    catch (InvalidTokenException e) {
      currentTokenStart = e.getOffset();
      reportInvalidToken(e);
    }
    catch (EmptyTokenException e) { }
  }

  /*
   * Make the replacement text for an entity out of the literal in the
   * current token.
   */
  private byte[] makeReplacementText(boolean allowPerefs) throws IOException {
    valueBuf.clear();
    Token t = new Token();
    int start = currentTokenStart + minBPC;
    final int end = bufStart - minBPC;
    try {
      for (;;) {
	int tok;
	int nextStart;
	try {
	  tok = enc.tokenizeEntityValue(buf, start, end, t);
	  nextStart = t.getTokenEnd();
	}
	catch (ExtensibleTokenException e) {
	  tok = e.getTokenType();
	  nextStart = end;
	}
	if (tok == Encoding.TOK_PARAM_ENTITY_REF && !allowPerefs) {
	  currentTokenStart = start;
	  fatal(MessageId.INTERNAL_PEREF_ENTVAL);
	}
	handleEntityValueToken(valueBuf, tok, start, nextStart, t);
	start = nextStart;
      }
    }
    catch (PartialTokenException e) {
      currentTokenStart = end;
      fatal(MessageId.NOT_WELL_FORMED);
    }
    catch (InvalidTokenException e) {
      currentTokenStart = e.getOffset();
      reportInvalidToken(e);
    }
    catch (EmptyTokenException e) { }

    return valueBuf.getBytes();
  }

  private void parseEntityValue(Buffer value) throws IOException {
    final Token t = new Token();
    for (;;) {
      int tok;
      for (;;) {
	try {
	  tok = enc.tokenizeEntityValue(buf, bufStart, bufEnd, t);
	  currentTokenStart = bufStart;
	  bufStart = t.getTokenEnd();
	  break;
	}
	catch (EmptyTokenException e) {
	  if (!fill())
	    return;
	}
	catch (PartialTokenException e) {
	  if (!fill()) {
	    currentTokenStart = bufStart;
	    bufStart = bufEnd;
	    fatal(MessageId.UNCLOSED_TOKEN);
	  }
	}
	catch (ExtensibleTokenException e) {
	  if (!fill()) {
	    currentTokenStart = bufStart;
	    bufStart = bufEnd;
	    tok = e.getTokenType();
	    break;
	  }
	}
	catch (InvalidTokenException e) {
	  currentTokenStart = e.getOffset();
	  reportInvalidToken(e);
	}
      }
      handleEntityValueToken(value, tok, currentTokenStart, bufStart, t);
    }
  }

  private void handleEntityValueToken(Buffer value, int tok, int start, int end, Token t) throws IOException {
    switch (tok) {
    case Encoding.TOK_DATA_CHARS:
    case Encoding.TOK_ENTITY_REF:
    case Encoding.TOK_MAGIC_ENTITY_REF:
      value.append(enc, buf, start, end);
      break;
    case Encoding.TOK_CHAR_REF:
      value.append(t.getRefChar());
      break;
    case Encoding.TOK_CHAR_PAIR_REF:
      value.appendRefCharPair(t);
      break;
    case Encoding.TOK_DATA_NEWLINE:
      value.append('\n');
      break;
    case Encoding.TOK_PARAM_ENTITY_REF:
      String name = stringCache.convert(buf,
					start + minBPC,
					end - minBPC,
					true);
      Entity entity = (Entity)dtd.paramEntityTable.get(name);
      if (entity == null) {
	if (dtd.complete)
	  fatal(MessageId.UNDEF_PEREF, name);
	break;
      }
      EntityParser parser = makeParserForEntity(entity, name, true);
      if (parser != null) {
	entity.open = true;
	parser.parseEntityValue(value);
	entity.open = false;
      }
      break;
    default:
      throw new Error("replacement text botch");
    }
  }

  private void parseMisc() throws IOException {
    try {
      for (;;) {
	switch (tokenizeProlog()) {
	case Encoding.TOK_PI:
	  nameStart = currentTokenStart + minBPC + minBPC;
	  app.processingInstruction(this);
	  break;
	case Encoding.TOK_COMMENT:
	case Encoding.TOK_PROLOG_S:
	  break;
	default:
	  fatal(MessageId.EPILOG_JUNK);
	}
      }
    }
    catch (EndOfPrologException e) {
      currentTokenStart = bufStart;
      fatal(MessageId.ELEMENT_AFTER_DOCUMENT_ELEMENT);
    }
    catch (EmptyTokenException e) { }
  }

  private final int tokenizeProlog()
       throws IOException, EmptyTokenException, EndOfPrologException {
    for (;;) {
      try {
	int tok = enc.tokenizeProlog(buf, bufStart, bufEnd, this);
	currentTokenStart = bufStart;
	bufStart = getTokenEnd();
	return tok;
      }
      catch (EmptyTokenException e) {
	if (!fill())
	  throw e;
      }
      catch (PartialTokenException e) {
	if (!fill()) {
	  currentTokenStart = bufStart;
	  bufStart = bufEnd;
	  fatal(MessageId.UNCLOSED_TOKEN);
	}
      }
      catch (ExtensibleTokenException e) {
	if (!fill()) {
	  currentTokenStart = bufStart;
	  bufStart = bufEnd;
	  return e.getTokenType();
	}
      }
      catch (InvalidTokenException e) {
	bufStart = currentTokenStart = e.getOffset();
	reportInvalidToken(e);
      }
    }
  }

  private static final int[] grow(int[] v) {
    int[] tem = v;
    v = new int[tem.length << 1];
    System.arraycopy(tem, 0, v, 0, tem.length);
    return v;
  }

  private long getEntityByteIndex(int off) {
    return bufEndStreamOffset - (bufEnd - off);
  }

  /* The size of the buffer is always a multiple of READSIZE.
     We do reads so that a complete read would end at the
     end of the buffer.  Unless there has been an incomplete
     read, we always read in multiples of READSIZE. */
  private boolean fill() throws IOException {
    if (in == null)
      return false;
    if (bufEnd == buf.length) {
      enc.movePosition(buf, posOff, bufStart, pos);
      /* The last read was complete. */
      int keep = bufEnd - bufStart;
      if (keep == 0)
	bufEnd = 0;
      else if (keep + READSIZE <= buf.length) {
	/*
	 * There is space in the buffer for at least READSIZE bytes.
	 * Choose bufEnd so that it is the least non-negative integer
	 * greater than or equal to <code>keep</code>, such
	 * <code>bufLength - keep</code> is a multiple of READSIZE.
	 */
	bufEnd = buf.length - (((buf.length - keep)/READSIZE) * READSIZE);
	for (int i = 0; i < keep; i++)
	  buf[bufEnd - keep + i] = buf[bufStart + i];
      }
      else {
	byte newBuf[] = new byte[buf.length + READSIZE];
	bufEnd = newBuf.length - READSIZE;
	for (int i = 0; i < keep; i++)
	  newBuf[bufEnd - keep + i] = buf[bufStart + i];
	buf = newBuf;
      }
      bufStart = bufEnd - keep;
      posOff = bufStart;
    }
    int nBytes = in.read(buf, bufEnd, buf.length - bufEnd);
    if (nBytes < 0) {
      in = null;
      return false;
    }
    bufEnd += nBytes;
    bufEndStreamOffset += nBytes;
    return true;
  }

  private void reportInvalidToken(InvalidTokenException e) throws NotWellFormedException {
    switch (e.getType()) {
    case InvalidTokenException.DUPLICATE_ATTRIBUTE:
      fatal(MessageId.DUPLICATE_ATTRIBUTE);
    case InvalidTokenException.XML_TARGET:
      fatal(MessageId.XML_TARGET);
    }
    fatal(MessageId.ILLEGAL_CHAR);
  }

  private void fatal(String message) throws NotWellFormedException {
    doFatal(message, null);
  }

  private void fatal(String message, Object arg) throws NotWellFormedException {
    doFatal(message, new Object[] { arg });
  }

  private void fatal(String message, Object arg1, Object arg2) throws NotWellFormedException {
    doFatal(message, new Object[] { arg1, arg2 });
  }

  private void doFatal(String id, Object[] args) throws NotWellFormedException {
    if (parent != null)
      parent.doFatal(id, args);
    if (posOff > currentTokenStart)
      throw new Error("positioning botch");
    if (enc != null)
      enc.movePosition(buf, posOff, currentTokenStart, pos);
    posOff = currentTokenStart;
    String desc = id;
    String message = null;
    try {
      ResourceBundle resources
	= ResourceBundle.getBundle("com.jclark.xml.parse.Messages", locale);
      desc = resources.getString(id);
      if (args != null)
	desc = MessageFormat.format(desc, args);
      Object[] msgArgs = new Object[] { desc,
					  location,
					  new Integer(pos.getLineNumber()),
					  new Integer(pos.getColumnNumber()),
					  new Long(getEntityByteIndex(currentTokenStart))
					  };
      message = MessageFormat.format(resources
				     .getString(MessageId.MESSAGE_FORMAT),
				     msgArgs);
    }
    catch (MissingResourceException e) { 
      message = desc;
    }
    catch (IllegalArgumentException e) {
      message = desc;
    }
    throw new NotWellFormedException(message,
				     desc,
				     location,
				     pos.getLineNumber(),
				     pos.getColumnNumber(),
				     getEntityByteIndex(currentTokenStart));
  }

  private static final
  boolean bytesEqual(byte[] buf1, int start1, int end1,
		     byte[] buf2, int start2, int end2) {
    int len = end1 - start1;
    if (end2 - start2 != len)
      return false;
    for (; len > 0; --len)
      if (buf1[start1++] != buf2[start2++])
	return false;
    return true;
  }

  private final static
  void copyBytes(byte[] from, int fromOff, byte[] to, int toOff, int len) {
    while (--len >= 0) {
      to[toOff++] = from[fromOff++];
    }
  }

  public String getName() {
    return stringCache.convert(buf, nameStart, getNameEnd(), true);
  }

  public DTD getDTD() {
    return dtd;
  }

  public int getLength() {
    if (data == null) {
      if (fixBPC != 0)
	return (getTokenEnd() - bufStart)/fixBPC;
      convertData(bufStart, getTokenEnd());
    }
    return dataLength;
  }

  public int getLengthMax() {
    if (data != null)
      return dataLength;
    else
      return (getTokenEnd() - bufStart)/minBPC;
  }

  public int copyChars(char[] cbuf, int off) {
    if (data != null) {
      System.arraycopy(data, 0, cbuf, off, dataLength);
      return dataLength;
    }
    else
      return enc.convert(buf, bufStart, getTokenEnd(), cbuf, off);
  }

  public void writeChars(Writer writer) throws IOException {
    if (data == null)
      convertData(bufStart, getTokenEnd());
    writer.write(data, 0, dataLength);
  }

  private void convertData(int start, int end) {
    if (dataBuf == null || dataBuf.length * minBPC < end - start)
      dataBuf = new char[(end - start)/minBPC];
    dataLength = enc.convert(buf, start, end, dataBuf, 0);
    data = dataBuf;
  }

  private final void setAttributeValue(int index, String value) {
    if (attValues == null)
      attValues = new String[index + 10];
    else if (index >= attValues.length) {
      String[] tem = new String[index << 1];
      System.arraycopy(attValues, 0, tem, 0, attValues.length);
      attValues = tem;
    }
    attValues[index] = value;
  }

  public final int getAttributeCount() {
    if (nAttributes < 0)
      buildAttributes();
    return nAttributes;
  }
  
  public final String getAttributeName(int i) {
    if (nAttributes < 0)
      buildAttributes();
    if (i >= nAttributes)
      throw new IndexOutOfBoundsException();
    return attNames[i];
  }

  public final String getAttributeValue(int i) {
    if (nAttributes < 0)
      buildAttributes();
    if (i < getAttributeSpecifiedCount()) {
      if (isAttributeNormalized(i))
	return stringCache.convert(buf,
				   getAttributeValueStart(i),
				   getAttributeValueEnd(i),
				   false);
    }
    else if (i >= nAttributes)
      throw new IndexOutOfBoundsException();
    return attValues[i];
  }

  public final String getAttributeValue(String name) {
    if (nAttributes < 0)
      buildAttributes();
    for (int i = 0; i < nAttributes; i++) {
      if (attNames[i] == name) {
	if (i < getAttributeSpecifiedCount() && isAttributeNormalized(i))
	  return stringCache.convert(buf,
				     getAttributeValueStart(i),
				     getAttributeValueEnd(i),
				     false);
	else
	  return attValues[i];
      }
    }
    return null;
  }

  private void buildAttributes() {
    ElementType elementType
      = (ElementType)dtd.elementTypeTable.get(getName());
    int nSpecAtts = getAttributeSpecifiedCount();

    {
      int totalAtts = nSpecAtts;
      if (elementType != null)
	totalAtts += elementType.getDefaultAttributeCount();
      if (attNames == null || totalAtts > attNames.length)
	attNames = new String[totalAtts];
    }
    for (int i = nSpecAtts; --i >= 0;)
      attNames[i] = stringCache.convert(buf,
					getAttributeNameStart(i),
					getAttributeNameEnd(i),
					true);
    nAttributes = nSpecAtts;
    if (elementType != null) {
      int nDefaults = elementType.getDefaultAttributeCount();
      if (defaultSpecified == null
	  || nDefaults > defaultSpecified.length)
	defaultSpecified = new boolean[nDefaults];
      else {
	for (int i = 0; i < nDefaults; i++)
	  defaultSpecified[i] = false;
      }
      for (int i = nSpecAtts; --i >= 0;) {
	int di = elementType.getAttributeDefaultIndex(attNames[i]);
	if (di >= 0)
	  defaultSpecified[di] = true;
      }
      for (int i = 0; i < nDefaults; i++) {
	if (!defaultSpecified[i]) {
	  setAttributeValue(nAttributes,
			    elementType.getDefaultAttributeValue(i));
	  attNames[nAttributes] = elementType.getDefaultAttributeName(i);
	  ++nAttributes;
	}
      }
    }
  }

  public final String getInstruction() {
    return normalizeNewlines(stringCache.convert(buf,
						 enc.skipS(buf,
							   getNameEnd(),
							   getTokenEnd()),
						 getTokenEnd() - 2*minBPC,
						 false));
  }

  private final String normalizeNewlines(String str) {
    int i = str.indexOf('\r');
    if (i < 0)
      return str;
    StringBuffer buf = new StringBuffer();
    for (i = 0; i < str.length(); i++) {
      char c = str.charAt(i);
      if (c == '\r') {
	buf.append('\n');
	if (i < str.length() && str.charAt(i + 1) != '\r')
	  i++;
      }
      else
	buf.append(c);
    }
    return buf.toString();
  }
}
