package com.jclark.xml.sax;

import com.jclark.xml.parse.*;
import java.net.URL;
import java.io.IOException;

class Bridge implements org.xml.sax.AttributeMap, EntityManager, Application {
  private org.xml.sax.EntityHandler entityHandler;
  private org.xml.sax.DocumentHandler documentHandler;
  private StartElementEvent event;
  private int hintAttributeIndex;
  private static final int INIT_DATA_BUF_SIZE = 80;
  private char[] dataBuf = new char[INIT_DATA_BUF_SIZE];
  private int dataBufUsed = 0;
  private AttributeNameEnumeration attNameEnumerator
    = new AttributeNameEnumeration();
  private boolean attNameEnumeratorActive = false;

  public Bridge(org.xml.sax.DocumentHandler documentHandler,
		org.xml.sax.EntityHandler entityHandler) {
    if (documentHandler == null)
      this.documentHandler = new org.xml.sax.HandlerBase();
    else
      this.documentHandler = documentHandler;
    this.entityHandler = entityHandler;
  }

  public OpenEntity open(String systemId, URL baseURL, String publicId) throws IOException {
    URL specURL = new URL(baseURL, systemId);
    URL useURL;
    if (entityHandler == null)
      useURL = specURL;
    else {
      String useSystemId;
      try {
	useSystemId = entityHandler.resolveEntity(null,
						  publicId,
						  specURL.toString());
	if (useSystemId == null)
	  return null;
      }
      catch (Exception e) {
	throw new WrapperException(e);
      }
      useURL = new URL(useSystemId);
    }
    return new OpenEntity(useURL.openStream(),
			  useURL.toString(),
			  specURL,
			  null);
  }

  public void startDocument() {
    try {
      documentHandler.startDocument();
    }
    catch (Exception e) {
      throw new WrapperException(e);
    }
  }

  public void startElement(StartElementEvent event) {
    flushData();
    try {
      this.event = event;
      documentHandler.startElement(event.getName(), this);
      attNameEnumeratorActive = false;
    }
    catch (Exception e) {
      throw new WrapperException(e);
    }
  }

  public void characterData(CharacterDataEvent event) {
    int need = event.getLengthMax() + dataBufUsed;
    if (need > dataBuf.length) {
      int newLength = dataBufUsed << 1;
      while (need > newLength)
	newLength <<= 1;
      char[] tem = dataBuf;
      dataBuf = new char[newLength];
      if (dataBufUsed > 0)
	System.arraycopy(tem, 0, dataBuf, 0, dataBufUsed);
    }
    dataBufUsed += event.copyChars(dataBuf, dataBufUsed);
  }

  private final void flushData() {
    if (dataBufUsed > 0) {
      try {
	documentHandler.characters(dataBuf, 0, dataBufUsed);
      }
      catch (Exception e) {
	throw new WrapperException(e);
      }
      dataBufUsed = 0;
    }
      
  }
  public void endElement(EndElementEvent event) {
    flushData();
    try {
      documentHandler.endElement(event.getName());
    }
    catch (Exception e) {
      throw new WrapperException(e);
    }
  }

  public void processingInstruction(ProcessingInstructionEvent event) {
    flushData();
    try {
      documentHandler.processingInstruction(event.getName(),
					    event.getInstruction());
    }
    catch (Exception e) {
      throw new WrapperException(e);
    }
  }

  public void endProlog(EndPrologEvent event) { }

  public void endDocument() {
    flushData();
    try {
      documentHandler.endDocument();
    }
    catch (Exception e) {
      throw new WrapperException(e);
    }
  }

  class AttributeNameEnumeration implements java.util.Enumeration {
    int i = 0;
    final void reset() { i = 0; }
    public Object nextElement() throws java.util.NoSuchElementException {
      try {
	hintAttributeIndex = i;
	return event.getAttributeName(i++);
      }
      catch (IndexOutOfBoundsException e) {
	throw new java.util.NoSuchElementException();
      }
    }
    public boolean hasMoreElements() {
      return i < event.getAttributeCount();
    }
  }

  public java.util.Enumeration getAttributeNames() {
    if (!attNameEnumeratorActive) {
      attNameEnumerator.reset();
      attNameEnumeratorActive = true;
      return attNameEnumerator;
    }
    return new AttributeNameEnumeration();
  }

  public String getValue(String attributeName) {
    int n = event.getAttributeCount();
    if (hintAttributeIndex < n
	&& attributeName == event.getAttributeName(hintAttributeIndex))
      return event.getAttributeValue(hintAttributeIndex);
    return event.getAttributeValue(attributeName);
  }

  public boolean isEntity(String aname) { return false; }
  public boolean isNotation(String aname) { return false; }
  public boolean isId(String aname) { return false; }
  public boolean isIdref(String aname) { return false; }
  public String getEntityPublicID(String aname) { return null; }
  public String getEntitySystemID(String aname) { return null; }
  public String getNotationName(String aname) { return null; }
  public String getNotationPublicID(String aname) { return null; }
  public String getNotationSystemID(String aname) { return null; }
}
